@@ -78,7 +78,7 @@ namespace {
   Creates an empty QQuickTextNode
 */
 QQuickTextNode::QQuickTextNode(QQuickItem *ownerElement)
-    : m_cursorNode(0), m_ownerElement(ownerElement), m_useNativeRenderer(false)
+    : m_cursorNode(nullptr), m_ownerElement(ownerElement), m_useNativeRenderer(false)
 {
 #ifdef QSG_RUNTIME_DESCRIPTION
     qsgnode_set_description(this, QLatin1String("text"));
@@ -125,16 +125,32 @@ QSGGlyphNode *QQuickTextNode::addGlyphs(const QPointF &position, const QGlyphRun
     node->geometry()->setIndexDataPattern(QSGGeometry::StaticPattern);
     node->geometry()->setVertexDataPattern(QSGGeometry::StaticPattern);
 
-    if (parentNode == 0)
+    if (parentNode == nullptr)
         parentNode = this;
     parentNode->appendChildNode(node);
 
+    if (style == QQuickText::Outline && color.alpha() > 0 && styleColor != color) {
+        QSGGlyphNode *fillNode = sg->sceneGraphContext()->createGlyphNode(sg, preferNativeGlyphNode);
+        fillNode->setOwnerElement(m_ownerElement);
+        fillNode->setGlyphs(position + QPointF(0, glyphs.rawFont().ascent()), glyphs);
+        fillNode->setStyle(QQuickText::Normal);
+        fillNode->setPreferredAntialiasingMode(QSGGlyphNode::GrayAntialiasing);
+        fillNode->setColor(color);
+        fillNode->update();
+
+        fillNode->geometry()->setIndexDataPattern(QSGGeometry::StaticPattern);
+        fillNode->geometry()->setVertexDataPattern(QSGGeometry::StaticPattern);
+
+        parentNode->appendChildNode(fillNode);
+        fillNode->setRenderOrder(node->renderOrder() + 1);
+    }
+
     return node;
 }
 
 void QQuickTextNode::setCursor(const QRectF &rect, const QColor &color)
 {
-    if (m_cursorNode != 0)
+    if (m_cursorNode != nullptr)
         delete m_cursorNode;
 
     QSGRenderContext *sg = QQuickItemPrivate::get(m_ownerElement)->sceneGraphRenderContext();
@@ -147,7 +163,7 @@ void QQuickTextNode::clearCursor()
     if (m_cursorNode)
         removeChildNode(m_cursorNode);
     delete m_cursorNode;
-    m_cursorNode = 0;
+    m_cursorNode = nullptr;
 }
 
 void QQuickTextNode::addRectangleNode(const QRectF &rect, const QColor &color)
@@ -160,7 +176,7 @@ void QQuickTextNode::addRectangleNode(const QRectF &rect, const QColor &color)
 void QQuickTextNode::addImage(const QRectF &rect, const QImage &image)
 {
     QSGRenderContext *sg = QQuickItemPrivate::get(m_ownerElement)->sceneGraphRenderContext();
-    QSGInternalImageNode *node = sg->sceneGraphContext()->createInternalImageNode();
+    QSGInternalImageNode *node = sg->sceneGraphContext()->createInternalImageNode(sg);
     QSGTexture *texture = sg->createTexture(image);
     if (m_ownerElement->smooth())
         texture->setFiltering(QSGTexture::Linear);
@@ -205,7 +221,7 @@ void QQuickTextNode::addTextDocument(const QPointF &position, QTextDocument *tex
 
             QTextBlock block = textFrame->firstCursorPosition().block();
             engine.setCurrentLine(block.layout()->lineForTextPosition(pos - block.position()));
-            engine.addTextObject(rect.topLeft(), format, QQuickTextNodeEngine::Unselected, textDocument,
+            engine.addTextObject(block, rect.topLeft(), format, QQuickTextNodeEngine::Unselected, textDocument,
                                  pos, textFrame->frameFormat().position());
         } else {
             QTextFrame::iterator it = textFrame->begin();
@@ -237,7 +253,7 @@ void QQuickTextNode::addTextLayout(const QPointF &position, QTextLayout *textLay
     engine.setAnchorColor(anchorColor);
     engine.setPosition(position);
 
-#ifndef QT_NO_IM
+#if QT_CONFIG(im)
     int preeditLength = textLayout->preeditAreaText().length();
     int preeditPosition = textLayout->preeditAreaPosition();
 #endif
@@ -256,7 +272,7 @@ void QQuickTextNode::addTextLayout(const QPointF &position, QTextLayout *textLay
         int length = line.textLength();
         int end = start + length;
 
-#ifndef QT_NO_IM
+#if QT_CONFIG(im)
         if (preeditPosition >= 0
          && preeditPosition >= start
          && preeditPosition < end) {
@@ -273,9 +289,9 @@ void QQuickTextNode::addTextLayout(const QPointF &position, QTextLayout *textLay
 
 void QQuickTextNode::deleteContent()
 {
-    while (firstChild() != 0)
+    while (firstChild() != nullptr)
         delete firstChild();
-    m_cursorNode = 0;
+    m_cursorNode = nullptr;
     qDeleteAll(m_textures);
     m_textures.clear();
 }
