@@ -2,10 +2,10 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-info: "StringLiteral :: 'SingleStringCharacters_opt'"
-es5id: 7.8.4_A1.2_T2
+info: "StringLiteral :: \"DoubleStringCharacters_opt\""
+es5id: 7.8.4_A1.1_T2
 description: >
-    SingleStringCharacter :: SourceCharacter but not single-quote ' or
+    DoubleStringCharacter :: SourceCharacter but not double-quote " or
     LineTerminator
 negative:
   phase: parse
@@ -15,5 +15,5 @@ negative:
 throw "Test262: This statement should not be evaluated.";
 
 //CHECK#1
-'
-'
+"
+"
