@@ -30,10 +30,11 @@
 
 #include <private/qv4compileddata_p.h>
 #include <private/qv4compiler_p.h>
-#include <private/qv4jsir_p.h>
-#include <private/qv4isel_p.h>
-#include <private/qv8engine_p.h>
 #include <private/qv4engine_p.h>
+#include <private/qv4codegen_p.h>
+#include <private/qqmlcomponent_p.h>
+#include <private/qv4executablecompilationunit_p.h>
+#include <private/qqmlscriptdata_p.h>
 #include <QQmlComponent>
 #include <QQmlEngine>
 #include <QQmlFileSelector>
@@ -48,7 +49,9 @@ class tst_qmldiskcache: public QObject
 
 private slots:
     void initTestCase();
+    void cleanupTestCase();
 
+    void loadLocalAsFallback();
     void regenerateAfterChange();
     void registerImportForImplicitComponent();
     void basicVersionChecks();
@@ -60,6 +63,11 @@ private slots:
     void stableOrderOfDependentCompositeTypes();
     void singletonDependency();
     void cppRegisteredSingletonDependency();
+    void cacheModuleScripts();
+    void reuseStaticMappings();
+
+private:
+    QDir m_qmlCacheDirectory;
 };
 
 // A wrapper around QQmlComponent to ensure the temporary reference counts
@@ -80,7 +88,9 @@ public:
 
     void waitForLoad()
     {
-        QTRY_VERIFY(status() == QQmlComponent::Ready || status() == QQmlComponent::Error);
+        QTRY_VERIFY_WITH_TIMEOUT(
+                    status() == QQmlComponent::Ready || status() == QQmlComponent::Error,
+                    32768);
     }
 };
 
@@ -90,7 +100,13 @@ static void waitForFileSystem()
     // the newly written file has a modification date newer than an existing cache file, we must
     // wait.
     // Similar effects of lacking precision have been observed on some Linux systems.
-    QThread::sleep(1);
+    static const bool fsHasSubSecondResolution = []() {
+        QDateTime mtime = QFileInfo(QStandardPaths::writableLocation(QStandardPaths::CacheLocation)).lastModified();
+        // 1:1000 chance of a false negative
+        return mtime.toMSecsSinceEpoch() % 1000;
+    }();
+    if (!fsHasSubSecondResolution)
+        QThread::sleep(1);
 }
 
 struct TestCompiler
@@ -107,38 +123,48 @@ struct TestCompiler
     {
         closeMapping();
         testFilePath = baseDirectory + QStringLiteral("/test.qml");
-        cacheFilePath = baseDirectory + QStringLiteral("/test.qmlc");
+        cacheFilePath = QV4::ExecutableCompilationUnit::localCacheFilePath(
+                QUrl::fromLocalFile(testFilePath));
         mappedFile.setFileName(cacheFilePath);
     }
 
-    bool compile(const QByteArray &contents)
+    void reset()
     {
         closeMapping();
         engine->clearComponentCache();
-
         waitForFileSystem();
+    }
 
-        {
-            QFile f(testFilePath);
-            if (!f.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
-                lastErrorString = f.errorString();
-                return false;
-            }
-            if (f.write(contents) != contents.size()) {
-                lastErrorString = f.errorString();
-                return false;
-            }
+    bool writeTestFile(const QByteArray &contents)
+    {
+        QFile f(testFilePath);
+        if (!f.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
+            lastErrorString = f.errorString();
+            return false;
         }
+        if (f.write(contents) != contents.size()) {
+            lastErrorString = f.errorString();
+            return false;
+        }
+        return true;
+    }
 
+    bool loadTestFile()
+    {
         CleanlyLoadingComponent component(engine, testFilePath);
         if (!component.isReady()) {
             lastErrorString = component.errorString();
             return false;
         }
-
         return true;
     }
 
+    bool compile(const QByteArray &contents)
+    {
+        reset();
+        return writeTestFile(contents) && loadTestFile();
+    }
+
     const QV4::CompiledData::Unit *mapUnit()
     {
         if (!mappedFile.open(QIODevice::ReadOnly)) {
@@ -157,28 +183,62 @@ struct TestCompiler
     }
 
     typedef void (*HeaderTweakFunction)(QV4::CompiledData::Unit *header);
-    bool tweakHeader(HeaderTweakFunction function)
+    bool tweakHeader(HeaderTweakFunction function, const QString &newName)
     {
         closeMapping();
 
-        QFile f(cacheFilePath);
-        if (!f.open(QIODevice::ReadWrite))
+        const QString targetTestFilePath = tempDir.path() + "/" + newName;
+
+        {
+            QFile testFile(testFilePath);
+            if (!testFile.copy(targetTestFilePath))
+                return false;
+        }
+
+        const QString targetCacheFilePath = QV4::ExecutableCompilationUnit::localCacheFilePath(
+                    QUrl::fromLocalFile(targetTestFilePath));
+
+        QFile source(cacheFilePath);
+        if (!source.copy(targetCacheFilePath))
             return false;
+
+        if (!source.open(QIODevice::ReadOnly))
+            return false;
+
+        QFile target(targetCacheFilePath);
+        if (!target.open(QIODevice::WriteOnly))
+            return false;
+
         QV4::CompiledData::Unit header;
-        if (f.read(reinterpret_cast<char *>(&header), sizeof(header)) != sizeof(header))
+        if (source.read(reinterpret_cast<char *>(&header), sizeof(header)) != sizeof(header))
             return false;
         function(&header);
-        f.seek(0);
-        return f.write(reinterpret_cast<const char *>(&header), sizeof(header)) == sizeof(header);
+
+        return target.write(reinterpret_cast<const char *>(&header), sizeof(header))
+                == sizeof(header);
     }
 
-    bool verify()
+    bool verify(const QString &fileName = QString())
     {
-        QV4::ExecutionEngine *v4 = QV8Engine::getV4(engine);
-        QQmlRefPointer<QV4::CompiledData::CompilationUnit> unit = v4->iselFactory->createUnitForLoading();
-        return unit->loadFromDisk(QUrl::fromLocalFile(testFilePath), QFileInfo(testFilePath).lastModified(), v4->iselFactory.data(), &lastErrorString);
+        const QString path = fileName.isEmpty() ? testFilePath : tempDir.path() + "/" + fileName;
+
+        QQmlRefPointer<QV4::ExecutableCompilationUnit> unit
+                = QV4::ExecutableCompilationUnit::create();
+        return unit->loadFromDisk(QUrl::fromLocalFile(path),
+                                  QFileInfo(path).lastModified(), &lastErrorString);
     }
 
+    quintptr unitData()
+    {
+        QQmlRefPointer<QV4::ExecutableCompilationUnit> unit
+                = QV4::ExecutableCompilationUnit::create();
+        return unit->loadFromDisk(QUrl::fromLocalFile(testFilePath),
+                                  QFileInfo(testFilePath).lastModified(), &lastErrorString)
+                ? quintptr(unit->unitData())
+                : 0;
+    }
+
+
     void closeMapping()
     {
         if (currentMapping) {
@@ -188,10 +248,11 @@ struct TestCompiler
         mappedFile.close();
     }
 
-    void clearCache()
+    void clearCache(const QString &fileName = QString())
     {
+        const QString path = fileName.isEmpty() ? testFilePath : tempDir.path() + "/" + fileName;
         closeMapping();
-        QFile::remove(cacheFilePath);
+        QFile::remove(path);
     }
 
     QQmlEngine *engine;
@@ -206,6 +267,59 @@ struct TestCompiler
 void tst_qmldiskcache::initTestCase()
 {
     qputenv("QML_FORCE_DISK_CACHE", "1");
+    QStandardPaths::setTestModeEnabled(true);
+
+    const QString cacheDirectory = QStandardPaths::writableLocation(QStandardPaths::CacheLocation);
+    m_qmlCacheDirectory.setPath(cacheDirectory + QLatin1String("/qmlcache"));
+    if (m_qmlCacheDirectory.exists())
+        QVERIFY(m_qmlCacheDirectory.removeRecursively());
+    QVERIFY(QDir::root().mkpath(m_qmlCacheDirectory.absolutePath()));
+}
+
+void tst_qmldiskcache::cleanupTestCase()
+{
+    m_qmlCacheDirectory.removeRecursively();
+}
+
+void tst_qmldiskcache::loadLocalAsFallback()
+{
+    QQmlEngine engine;
+    TestCompiler testCompiler(&engine);
+
+    QVERIFY(testCompiler.tempDir.isValid());
+
+    const QByteArray contents = QByteArrayLiteral("import QtQml 2.0\n"
+                                                  "QtObject {\n"
+                                                   "    property string blah: Qt.platform;\n"
+                                                   "}");
+
+    QVERIFY2(testCompiler.compile(contents), qPrintable(testCompiler.lastErrorString));
+
+    // Create an invalid side-by-side .qmlc
+    {
+        QFile f(testCompiler.tempDir.path() + "/test.qmlc");
+        QVERIFY(f.open(QIODevice::WriteOnly | QIODevice::Truncate));
+        QV4::CompiledData::Unit unit = {};
+        memcpy(unit.magic, QV4::CompiledData::magic_str, sizeof(unit.magic));
+        unit.version = QV4_DATA_STRUCTURE_VERSION;
+        unit.qtVersion = QT_VERSION;
+        unit.sourceTimeStamp = testCompiler.mappedFile.fileTime(QFile::FileModificationTime).toMSecsSinceEpoch();
+        unit.unitSize = ~0U;    // make the size a silly number
+        // write something to the library hash that should cause it not to be loaded
+        memset(unit.libraryVersionHash, 'z', sizeof(unit.libraryVersionHash));
+        memset(unit.md5Checksum, 0, sizeof(unit.md5Checksum));
+
+        // leave the other fields unset, since they don't matter
+
+        f.write(reinterpret_cast<const char *>(&unit), sizeof(unit));
+    }
+
+    QQmlRefPointer<QV4::ExecutableCompilationUnit> unit = QV4::ExecutableCompilationUnit::create();
+    bool loaded = unit->loadFromDisk(QUrl::fromLocalFile(testCompiler.testFilePath),
+                                     QFileInfo(testCompiler.testFilePath).lastModified(),
+                                     &testCompiler.lastErrorString);
+    QVERIFY2(loaded, qPrintable(testCompiler.lastErrorString));
+    QCOMPARE(unit->objectCount(), 1);
 }
 
 void tst_qmldiskcache::regenerateAfterChange()
@@ -226,17 +340,21 @@ void tst_qmldiskcache::regenerateAfterChange()
         const QV4::CompiledData::Unit *testUnit = testCompiler.mapUnit();
         QVERIFY2(testUnit, qPrintable(testCompiler.lastErrorString));
 
-        QCOMPARE(quint32(testUnit->nObjects), quint32(1));
+        const QV4::CompiledData::QmlUnit *qmlUnit = testUnit->qmlUnit();
 
-        const QV4::CompiledData::Object *obj = testUnit->objectAt(0);
+        QCOMPARE(quint32(qmlUnit->nObjects), quint32(1));
+
+        const QV4::CompiledData::Object *obj = qmlUnit->objectAt(0);
         QCOMPARE(quint32(obj->nBindings), quint32(1));
         QCOMPARE(quint32(obj->bindingTable()->type), quint32(QV4::CompiledData::Binding::Type_Script));
-        QCOMPARE(quint32(obj->bindingTable()->value.compiledScriptIndex), quint32(1));
+        QCOMPARE(quint32(obj->bindingTable()->value.compiledScriptIndex), quint32(0));
 
-        QCOMPARE(quint32(testUnit->functionTableSize), quint32(2));
+        QCOMPARE(quint32(testUnit->functionTableSize), quint32(1));
 
-        const QV4::CompiledData::Function *bindingFunction = testUnit->functionAt(1);
-        QVERIFY(bindingFunction->codeOffset > testUnit->unitSize);
+        const QV4::CompiledData::Function *bindingFunction = testUnit->functionAt(0);
+        QCOMPARE(testUnit->stringAtInternal(bindingFunction->nameIndex), QString("expression for blah")); // check if we have the correct function
+        QVERIFY(bindingFunction->codeSize > 0);
+        QVERIFY(bindingFunction->codeOffset < testUnit->unitSize);
     }
 
     {
@@ -250,17 +368,24 @@ void tst_qmldiskcache::regenerateAfterChange()
         const QV4::CompiledData::Unit *testUnit = testCompiler.mapUnit();
         QVERIFY2(testUnit, qPrintable(testCompiler.lastErrorString));
 
-        QCOMPARE(quint32(testUnit->nObjects), quint32(1));
+        const QV4::CompiledData::QmlUnit *qmlUnit = testUnit->qmlUnit();
 
-        const QV4::CompiledData::Object *obj = testUnit->objectAt(0);
+        QCOMPARE(quint32(qmlUnit->nObjects), quint32(1));
+
+        const QV4::CompiledData::Object *obj = qmlUnit->objectAt(0);
         QCOMPARE(quint32(obj->nBindings), quint32(2));
         QCOMPARE(quint32(obj->bindingTable()->type), quint32(QV4::CompiledData::Binding::Type_Number));
-        QCOMPARE(obj->bindingTable()->valueAsNumber(), double(42));
 
-        QCOMPARE(quint32(testUnit->functionTableSize), quint32(2));
+        QCOMPARE(reinterpret_cast<const QV4::Value *>(testUnit->constants())
+                         [obj->bindingTable()->value.constantValueIndex].doubleValue(),
+                 double(42));
 
-        const QV4::CompiledData::Function *bindingFunction = testUnit->functionAt(1);
-        QVERIFY(bindingFunction->codeOffset > testUnit->unitSize);
+        QCOMPARE(quint32(testUnit->functionTableSize), quint32(1));
+
+        const QV4::CompiledData::Function *bindingFunction = testUnit->functionAt(0);
+        QCOMPARE(testUnit->stringAtInternal(bindingFunction->nameIndex), QString("expression for blah")); // check if we have the correct function
+        QVERIFY(bindingFunction->codeSize > 0);
+        QVERIFY(bindingFunction->codeOffset < testUnit->unitSize);
     }
 }
 
@@ -281,22 +406,23 @@ void tst_qmldiskcache::registerImportForImplicitComponent()
         const QV4::CompiledData::Unit *testUnit = testCompiler.mapUnit();
         QVERIFY2(testUnit, qPrintable(testCompiler.lastErrorString));
 
-        QCOMPARE(quint32(testUnit->nImports), quint32(2));
-        QCOMPARE(testUnit->stringAt(testUnit->importAt(0)->uriIndex), QStringLiteral("QtQuick"));
+        const QV4::CompiledData::QmlUnit *qmlUnit = testUnit->qmlUnit();
+        QCOMPARE(quint32(qmlUnit->nImports), quint32(2));
+        QCOMPARE(testUnit->stringAtInternal(qmlUnit->importAt(0)->uriIndex), QStringLiteral("QtQuick"));
 
         QQmlType componentType = QQmlMetaType::qmlType(&QQmlComponent::staticMetaObject);
 
-        QCOMPARE(testUnit->stringAt(testUnit->importAt(1)->uriIndex), QString(componentType.module()));
-        QCOMPARE(testUnit->stringAt(testUnit->importAt(1)->qualifierIndex), QStringLiteral("QmlInternals"));
+        QCOMPARE(testUnit->stringAtInternal(qmlUnit->importAt(1)->uriIndex), QString(componentType.module()));
+        QCOMPARE(testUnit->stringAtInternal(qmlUnit->importAt(1)->qualifierIndex), QStringLiteral("QmlInternals"));
 
-        QCOMPARE(quint32(testUnit->nObjects), quint32(3));
+        QCOMPARE(quint32(qmlUnit->nObjects), quint32(3));
 
-        const QV4::CompiledData::Object *obj = testUnit->objectAt(0);
+        const QV4::CompiledData::Object *obj = qmlUnit->objectAt(0);
         QCOMPARE(quint32(obj->nBindings), quint32(1));
         QCOMPARE(quint32(obj->bindingTable()->type), quint32(QV4::CompiledData::Binding::Type_Object));
 
-        const QV4::CompiledData::Object *implicitComponent = testUnit->objectAt(obj->bindingTable()->value.objectIndex);
-        QCOMPARE(testUnit->stringAt(implicitComponent->inheritedTypeNameIndex), QStringLiteral("QmlInternals.") + componentType.elementName());
+        const QV4::CompiledData::Object *implicitComponent = qmlUnit->objectAt(obj->bindingTable()->value.objectIndex);
+        QCOMPARE(testUnit->stringAtInternal(implicitComponent->inheritedTypeNameIndex), QStringLiteral("QmlInternals.") + componentType.elementName());
     }
 }
 
@@ -322,49 +448,28 @@ void tst_qmldiskcache::basicVersionChecks()
         testCompiler.clearCache();
         QVERIFY2(testCompiler.compile(contents), qPrintable(testCompiler.lastErrorString));
 
-        testCompiler.tweakHeader([](QV4::CompiledData::Unit *header) {
+        const QString qtVersionFile = QStringLiteral("qtversion.qml");
+        QVERIFY(testCompiler.tweakHeader([](QV4::CompiledData::Unit *header) {
             header->qtVersion = 0;
-        });
+        }, qtVersionFile));
 
-        QVERIFY(!testCompiler.verify());
+        QVERIFY(!testCompiler.verify(qtVersionFile));
         QCOMPARE(testCompiler.lastErrorString, QString::fromUtf8("Qt version mismatch. Found 0 expected %1").arg(QT_VERSION, 0, 16));
-        testCompiler.clearCache();
+        testCompiler.clearCache(qtVersionFile);
     }
 
     {
         testCompiler.clearCache();
         QVERIFY2(testCompiler.compile(contents), qPrintable(testCompiler.lastErrorString));
 
-        testCompiler.tweakHeader([](QV4::CompiledData::Unit *header) {
+        const QString versionFile = QStringLiteral("version.qml");
+        QVERIFY(testCompiler.tweakHeader([](QV4::CompiledData::Unit *header) {
             header->version = 0;
-        });
+        }, versionFile));
 
-        QVERIFY(!testCompiler.verify());
+        QVERIFY(!testCompiler.verify(versionFile));
         QCOMPARE(testCompiler.lastErrorString, QString::fromUtf8("V4 data structure version mismatch. Found 0 expected %1").arg(QV4_DATA_STRUCTURE_VERSION, 0, 16));
-    }
-
-    {
-        testCompiler.clearCache();
-        QVERIFY2(testCompiler.compile(contents), qPrintable(testCompiler.lastErrorString));
-
-        testCompiler.tweakHeader([](QV4::CompiledData::Unit *header) {
-            header->architectureIndex = 0;
-        });
-
-        QVERIFY(!testCompiler.verify());
-        QCOMPARE(testCompiler.lastErrorString, QString::fromUtf8("Architecture mismatch. Found  expected %1").arg(QSysInfo::buildAbi()));
-    }
-
-    {
-        testCompiler.clearCache();
-        QVERIFY2(testCompiler.compile(contents), qPrintable(testCompiler.lastErrorString));
-
-        testCompiler.tweakHeader([](QV4::CompiledData::Unit *header) {
-            header->codeGeneratorIndex = 0;
-        });
-
-        QVERIFY(!testCompiler.verify());
-        QCOMPARE(testCompiler.lastErrorString, QString::fromUtf8("Code generator mismatch. Found code generated by  but expected %1").arg(QV8Engine::getV4(&engine)->iselFactory->codeGeneratorName));
+        testCompiler.clearCache(versionFile);
     }
 }
 
@@ -469,6 +574,10 @@ void tst_qmldiskcache::recompileAfterDirectoryChange()
         testCompiler.clearCache();
         QVERIFY2(testCompiler.compile(contents), qPrintable(testCompiler.lastErrorString));
         QVERIFY2(testCompiler.verify(), qPrintable(testCompiler.lastErrorString));
+        const QV4::CompiledData::Unit *unit = testCompiler.mapUnit();
+        QVERIFY(unit->sourceFileIndex != 0);
+        const QString expectedPath = QUrl::fromLocalFile(testCompiler.testFilePath).toString();
+        QCOMPARE(unit->stringAtInternal(unit->sourceFileIndex), expectedPath);
         testCompiler.closeMapping();
     }
 
@@ -521,7 +630,8 @@ void tst_qmldiskcache::fileSelectors()
         QVERIFY(!obj.isNull());
         QCOMPARE(obj->property("value").toInt(), 42);
 
-        QFile cacheFile(testFilePath + "c");
+        QFile cacheFile(QV4::ExecutableCompilationUnit::localCacheFilePath(
+                QUrl::fromLocalFile(testFilePath)));
         QVERIFY2(cacheFile.exists(), qPrintable(cacheFile.fileName()));
     }
 
@@ -536,7 +646,8 @@ void tst_qmldiskcache::fileSelectors()
         QVERIFY(!obj.isNull());
         QCOMPARE(obj->property("value").toInt(), 100);
 
-        QFile cacheFile(selectedTestFilePath + "c");
+        QFile cacheFile(QV4::ExecutableCompilationUnit::localCacheFilePath(
+                QUrl::fromLocalFile(selectedTestFilePath)));
         QVERIFY2(cacheFile.exists(), qPrintable(cacheFile.fileName()));
     }
 }
@@ -584,16 +695,21 @@ void tst_qmldiskcache::localAliases()
     }
 }
 
+static QSet<QString> entrySet(const QDir &dir)
+{
+    const auto &list = dir.entryList(QDir::Files | QDir::NoDotAndDotDot);
+    return QSet<QString>(list.cbegin(), list.cend());
+}
+
+static QSet<QString> entrySet(const QDir &dir, const QStringList &filters)
+{
+    const auto &list = dir.entryList(filters);
+    return QSet<QString>(list.cbegin(), list.cend());
+}
+
 void tst_qmldiskcache::cacheResources()
 {
-    const QString cacheDirectory = QStandardPaths::writableLocation(QStandardPaths::CacheLocation);
-    QVERIFY(QDir::root().mkpath(cacheDirectory));
-
-    const QString qmlCacheDirectory = cacheDirectory + QLatin1String("/qmlcache/");
-    QVERIFY(QDir(qmlCacheDirectory).removeRecursively());
-    QVERIFY(QDir::root().mkpath(qmlCacheDirectory));
-    QVERIFY(QDir(qmlCacheDirectory).entryList(QDir::NoDotAndDotDot).isEmpty());
-
+    const QSet<QString> existingFiles = entrySet(m_qmlCacheDirectory);
 
     QQmlEngine engine;
 
@@ -604,13 +720,13 @@ void tst_qmldiskcache::cacheResources()
         QCOMPARE(obj->property("value").toInt(), 20);
     }
 
-    const QStringList entries = QDir(qmlCacheDirectory).entryList(QDir::NoDotAndDotDot | QDir::Files);
+    const QSet<QString> entries = entrySet(m_qmlCacheDirectory).subtract(existingFiles);
     QCOMPARE(entries.count(), 1);
 
     QDateTime cacheFileTimeStamp;
 
     {
-        QFile cacheFile(qmlCacheDirectory + QLatin1Char('/') + entries.constFirst());
+        QFile cacheFile(m_qmlCacheDirectory.absoluteFilePath(*entries.cbegin()));
         QVERIFY2(cacheFile.open(QIODevice::ReadOnly), qPrintable(cacheFile.errorString()));
         QV4::CompiledData::Unit unit;
         QVERIFY(cacheFile.read(reinterpret_cast<char *>(&unit), sizeof(unit)) == sizeof(unit));
@@ -633,10 +749,11 @@ void tst_qmldiskcache::cacheResources()
     }
 
     {
-        const QStringList entries = QDir(qmlCacheDirectory).entryList(QDir::NoDotAndDotDot | QDir::Files);
+        const QSet<QString> entries = entrySet(m_qmlCacheDirectory).subtract(existingFiles);
         QCOMPARE(entries.count(), 1);
 
-        QCOMPARE(QFileInfo(qmlCacheDirectory + QLatin1Char('/') + entries.constFirst()).lastModified().toMSecsSinceEpoch(), cacheFileTimeStamp.toMSecsSinceEpoch());
+        QCOMPARE(QFileInfo(m_qmlCacheDirectory.absoluteFilePath(*entries.cbegin())).lastModified().toMSecsSinceEpoch(),
+                           cacheFileTimeStamp.toMSecsSinceEpoch());
     }
 }
 
@@ -689,7 +806,8 @@ void tst_qmldiskcache::stableOrderOfDependentCompositeTypes()
     QVERIFY2(firstDependentTypeClassName.contains("QMLTYPE"), firstDependentTypeClassName.constData());
     QVERIFY2(secondDependentTypeClassName.contains("QMLTYPE"), secondDependentTypeClassName.constData());
 
-    const QString testFileCachePath = testFilePath + QLatin1Char('c');
+    const QString testFileCachePath = QV4::ExecutableCompilationUnit::localCacheFilePath(
+            QUrl::fromLocalFile(testFilePath));
     QVERIFY(QFile::exists(testFileCachePath));
     QDateTime initialCacheTimeStamp = QFileInfo(testFileCachePath).lastModified();
 
@@ -767,7 +885,8 @@ void tst_qmldiskcache::singletonDependency()
         QCOMPARE(obj->property("value").toInt(), 42);
     }
 
-    const QString testFileCachePath = testFilePath + QLatin1Char('c');
+    const QString testFileCachePath = QV4::ExecutableCompilationUnit::localCacheFilePath(
+            QUrl::fromLocalFile(testFilePath));
     QVERIFY(QFile::exists(testFileCachePath));
     QDateTime initialCacheTimeStamp = QFileInfo(testFileCachePath).lastModified();
 
@@ -824,7 +943,8 @@ void tst_qmldiskcache::cppRegisteredSingletonDependency()
         QCOMPARE(value.toInt(), 42);
     }
 
-    const QString testFileCachePath = testFilePath + QLatin1Char('c');
+    const QString testFileCachePath = QV4::ExecutableCompilationUnit::localCacheFilePath(
+            QUrl::fromLocalFile(testFilePath));
     QVERIFY(QFile::exists(testFileCachePath));
     QDateTime initialCacheTimeStamp = QFileInfo(testFileCachePath).lastModified();
 
@@ -851,6 +971,65 @@ void tst_qmldiskcache::cppRegisteredSingletonDependency()
     }
 }
 
+void tst_qmldiskcache::cacheModuleScripts()
+{
+    const QSet<QString> existingFiles = entrySet(m_qmlCacheDirectory);
+
+    QQmlEngine engine;
+
+    {
+        CleanlyLoadingComponent component(&engine, QUrl("qrc:/importmodule.qml"));
+        QScopedPointer<QObject> obj(component.create());
+        QVERIFY(!obj.isNull());
+        QVERIFY(obj->property("ok").toBool());
+
+        auto componentPrivate = QQmlComponentPrivate::get(&component);
+        QVERIFY(componentPrivate);
+        auto compilationUnit = componentPrivate->compilationUnit->dependentScripts.first()->compilationUnit();
+        QVERIFY(compilationUnit);
+        auto unitData = compilationUnit->unitData();
+        QVERIFY(unitData);
+        QVERIFY(unitData->flags & QV4::CompiledData::Unit::StaticData);
+        QVERIFY(unitData->flags & QV4::CompiledData::Unit::IsESModule);
+        QVERIFY(!compilationUnit->backingFile.isNull());
+    }
+
+    const QSet<QString> entries = entrySet(m_qmlCacheDirectory, QStringList("*.mjsc"));
+
+    QCOMPARE(entries.count(), 1);
+
+    QDateTime cacheFileTimeStamp;
+
+    {
+        QFile cacheFile(m_qmlCacheDirectory.absoluteFilePath(*entries.cbegin()));
+        QVERIFY2(cacheFile.open(QIODevice::ReadOnly), qPrintable(cacheFile.errorString()));
+        QV4::CompiledData::Unit unit;
+        QVERIFY(cacheFile.read(reinterpret_cast<char *>(&unit), sizeof(unit)) == sizeof(unit));
+
+        QVERIFY(unit.flags & QV4::CompiledData::Unit::IsESModule);
+    }
+}
+
+void tst_qmldiskcache::reuseStaticMappings()
+{
+    QQmlEngine engine;
+
+    TestCompiler testCompiler(&engine);
+    QVERIFY(testCompiler.tempDir.isValid());
+
+    QVERIFY2(testCompiler.compile("import QtQml 2.15\nQtObject { objectName: 'foobar' }\n"),
+             qPrintable(testCompiler.lastErrorString));
+
+    const quintptr data1 = testCompiler.unitData();
+    QVERIFY(data1 != 0);
+    QCOMPARE(testCompiler.unitData(), data1);
+
+    testCompiler.reset();
+    QVERIFY(testCompiler.loadTestFile());
+
+    QCOMPARE(testCompiler.unitData(), data1);
+}
+
 QTEST_MAIN(tst_qmldiskcache)
 
 #include "tst_qmldiskcache.moc"
