@@ -1,37 +1,33 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing/
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the test suite of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:GPL-EXCEPT$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
 ** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see http://www.qt.io/terms-conditions. For further
-** information use the contact form at http://www.qt.io/contact-us.
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** As a special exception, The Qt Company gives you certain additional
-** rights. These rights are described in The Qt Company LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
 ****************************************************************************/
 #include <qtest.h>
 #include <QtQml/qqmlengine.h>
+#include <QtQml/qqmlexpression.h>
 #include <QtQml/qqmlcomponent.h>
 #include <QtQuick/qquickview.h>
 #include <QtQuick/private/qquickrectangle_p.h>
@@ -45,6 +41,23 @@
 
 Q_DECLARE_METATYPE(QQuickImageBase::Status)
 
+template <typename T> static T evaluate(QObject *scope, const QString &expression)
+{
+    QQmlExpression expr(qmlContext(scope), scope, expression);
+    QVariant result = expr.evaluate();
+    if (expr.hasError())
+        qWarning() << expr.error().toString();
+    return result.value<T>();
+}
+
+template <> void evaluate<void>(QObject *scope, const QString &expression)
+{
+    QQmlExpression expr(qmlContext(scope), scope, expression);
+    expr.evaluate();
+    if (expr.hasError())
+        qWarning() << expr.error().toString();
+}
+
 class tst_qquickanimatedimage : public QQmlDataTest
 {
     Q_OBJECT
@@ -65,12 +78,15 @@ private slots:
     void remote_data();
     void sourceSize();
     void sourceSizeChanges();
+    void sourceSizeChanges_intermediate();
     void sourceSizeReadOnly();
     void invalidSource();
     void qtbug_16520();
     void progressAndStatusChanges();
     void playingAndPausedChanges();
     void noCaching();
+    void sourceChangesOnFrameChanged();
+    void currentFrame();
 };
 
 void tst_qquickanimatedimage::cleanup()
@@ -136,6 +152,16 @@ void tst_qquickanimatedimage::frameCount()
     QVERIFY(anim->isPlaying());
     QCOMPARE(anim->frameCount(), 3);
 
+    QSignalSpy frameCountChangedSpy(anim, &QQuickAnimatedImage::frameCountChanged);
+
+    const QUrl origSource = anim->source();
+    anim->setSource(QUrl());
+    QCOMPARE(anim->frameCount(), 0);
+    QCOMPARE(frameCountChangedSpy.count(), 1);
+    anim->setSource(origSource);
+    QCOMPARE(anim->frameCount(), 3);
+    QCOMPARE(frameCountChangedSpy.count(), 2);
+
     delete anim;
 }
 
@@ -146,7 +172,7 @@ void tst_qquickanimatedimage::mirror_running()
     QQuickView window;
     window.setSource(testFileUrl("hearts.qml"));
     window.show();
-    QTest::qWaitForWindowExposed(&window);
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
 
     QQuickAnimatedImage *anim = qobject_cast<QQuickAnimatedImage *>(window.rootObject());
     QVERIFY(anim);
@@ -183,7 +209,11 @@ void tst_qquickanimatedimage::mirror_running()
     QImage frame0_expected = frame0.transformed(transform);
     QImage frame1_expected = frame1.transformed(transform);
 
+    if (window.devicePixelRatio() != 1.0 && window.rendererInterface()->graphicsApi() == QSGRendererInterface::Software)
+        QSKIP("QTBUG-53823");
     QCOMPARE(frame0_flipped, frame0_expected);
+    if (window.devicePixelRatio() != 1.0 && window.rendererInterface()->graphicsApi() == QSGRendererInterface::Software)
+        QSKIP("QTBUG-53823");
     QCOMPARE(frame1_flipped, frame1_expected);
 
     delete anim;
@@ -210,16 +240,14 @@ void tst_qquickanimatedimage::mirror_notRunning()
 
     int frame = anim->currentFrame();
     bool playing = anim->isPlaying();
-    bool paused = anim->isPlaying();
+    bool paused = anim->isPaused();
 
     anim->setProperty("mirror", true);
     screenshot = window.grabWindow();
 
     screenshot.save("screen.png");
-#if defined(Q_OS_WIN)
-    // QTBUG-36717
-    QSKIP("This test is failing in the CI system under mysterious circumstances");
-#endif
+    if (window.devicePixelRatio() != 1.0 && window.rendererInterface()->graphicsApi() == QSGRendererInterface::Software)
+        QSKIP("QTBUG-53823");
     QCOMPARE(screenshot, expected);
 
     // mirroring should not change the current frame or playing status
@@ -252,9 +280,7 @@ void tst_qquickanimatedimage::remote()
     QFETCH(QString, fileName);
     QFETCH(bool, paused);
 
-    TestHTTPServer server;
-    QVERIFY2(server.listen(), qPrintable(server.errorString()));
-    server.serveDirectory(dataDirectory());
+    ThreadedTestHTTPServer server(dataDirectory());
 
     QQmlEngine engine;
     QQmlComponent component(&engine, server.url(fileName));
@@ -328,59 +354,81 @@ void tst_qquickanimatedimage::sourceSizeChanges()
     QQmlContext *ctxt = engine.rootContext();
     ctxt->setContextProperty("srcImage", "");
     QQuickAnimatedImage *anim = qobject_cast<QQuickAnimatedImage*>(component.create());
-    QVERIFY(anim != 0);
+    QVERIFY(anim != nullptr);
 
     QSignalSpy sourceSizeSpy(anim, SIGNAL(sourceSizeChanged()));
 
     // Local
     ctxt->setContextProperty("srcImage", QUrl(""));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Null);
-    QTRY_VERIFY(sourceSizeSpy.count() == 0);
+    QTRY_COMPARE(sourceSizeSpy.count(), 0);
 
     ctxt->setContextProperty("srcImage", testFileUrl("hearts.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 1);
+    QTRY_COMPARE(sourceSizeSpy.count(), 1);
 
     ctxt->setContextProperty("srcImage", testFileUrl("hearts.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 1);
+    QTRY_COMPARE(sourceSizeSpy.count(), 1);
 
     ctxt->setContextProperty("srcImage", testFileUrl("hearts_copy.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 1);
+    QTRY_COMPARE(sourceSizeSpy.count(), 1);
 
     ctxt->setContextProperty("srcImage", testFileUrl("colors.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 2);
+    QTRY_COMPARE(sourceSizeSpy.count(), 2);
 
     ctxt->setContextProperty("srcImage", QUrl(""));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Null);
-    QTRY_VERIFY(sourceSizeSpy.count() == 3);
+    QTRY_COMPARE(sourceSizeSpy.count(), 3);
 
     // Remote
     ctxt->setContextProperty("srcImage", server.url("/hearts.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 4);
+    QTRY_COMPARE(sourceSizeSpy.count(), 4);
 
     ctxt->setContextProperty("srcImage", server.url("/hearts.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 4);
+    QTRY_COMPARE(sourceSizeSpy.count(), 4);
 
     ctxt->setContextProperty("srcImage", server.url("/hearts_copy.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 4);
+    QTRY_COMPARE(sourceSizeSpy.count(), 4);
 
     ctxt->setContextProperty("srcImage", server.url("/colors.gif"));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
-    QTRY_VERIFY(sourceSizeSpy.count() == 5);
+    QTRY_COMPARE(sourceSizeSpy.count(), 5);
 
     ctxt->setContextProperty("srcImage", QUrl(""));
     QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Null);
-    QTRY_VERIFY(sourceSizeSpy.count() == 6);
+    QTRY_COMPARE(sourceSizeSpy.count(), 6);
 
     delete anim;
 }
 
+void tst_qquickanimatedimage::sourceSizeChanges_intermediate()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine);
+    component.setData("import QtQuick 2.0\nAnimatedImage { readonly property int testWidth: status === AnimatedImage.Ready ? sourceSize.width : -1; source: srcImage }", QUrl::fromLocalFile(""));
+    QTRY_VERIFY(component.isReady());
+    QQmlContext *ctxt = engine.rootContext();
+    ctxt->setContextProperty("srcImage", "");
+
+    QScopedPointer<QQuickAnimatedImage> anim(qobject_cast<QQuickAnimatedImage*>(component.create()));
+    QVERIFY(anim != nullptr);
+
+    ctxt->setContextProperty("srcImage", testFileUrl("hearts.gif"));
+    QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
+    QTRY_COMPARE(anim->property("testWidth").toInt(), anim->sourceSize().width());
+
+    ctxt->setContextProperty("srcImage", testFileUrl("hearts_copy.gif"));
+    QTRY_COMPARE(anim->status(), QQuickAnimatedImage::Ready);
+    QTRY_COMPARE(anim->property("testWidth").toInt(), anim->sourceSize().width());
+}
+
+
 void tst_qquickanimatedimage::qtbug_16520()
 {
     TestHTTPServer server;
@@ -394,11 +442,11 @@ void tst_qquickanimatedimage::qtbug_16520()
     QQuickRectangle *root = qobject_cast<QQuickRectangle *>(component.create());
     QVERIFY(root);
     QQuickAnimatedImage *anim = root->findChild<QQuickAnimatedImage*>("anim");
-    QVERIFY(anim != 0);
+    QVERIFY(anim != nullptr);
 
     anim->setProperty("source", server.urlString("/stickman.gif"));
-    QTRY_VERIFY(anim->opacity() == 0);
-    QTRY_VERIFY(anim->opacity() == 1);
+    QTRY_COMPARE(anim->opacity(), qreal(0));
+    QTRY_COMPARE(anim->opacity(), qreal(1));
 
     delete anim;
     delete root;
@@ -417,9 +465,9 @@ void tst_qquickanimatedimage::progressAndStatusChanges()
     QQmlComponent component(&engine);
     component.setData(componentStr.toLatin1(), QUrl::fromLocalFile(""));
     QQuickImage *obj = qobject_cast<QQuickImage*>(component.create());
-    QVERIFY(obj != 0);
-    QVERIFY(obj->status() == QQuickImage::Ready);
-    QTRY_VERIFY(obj->progress() == 1.0);
+    QVERIFY(obj != nullptr);
+    QCOMPARE(obj->status(), QQuickImage::Ready);
+    QTRY_COMPARE(obj->progress(), 1.0);
 
     qRegisterMetaType<QQuickImageBase::Status>();
     QSignalSpy sourceSpy(obj, SIGNAL(sourceChanged(QUrl)));
@@ -428,33 +476,33 @@ void tst_qquickanimatedimage::progressAndStatusChanges()
 
     // Same image
     ctxt->setContextProperty("srcImage", testFileUrl("stickman.gif"));
-    QTRY_VERIFY(obj->status() == QQuickImage::Ready);
-    QTRY_VERIFY(obj->progress() == 1.0);
+    QTRY_COMPARE(obj->status(), QQuickImage::Ready);
+    QTRY_COMPARE(obj->progress(), 1.0);
     QTRY_COMPARE(sourceSpy.count(), 0);
     QTRY_COMPARE(progressSpy.count(), 0);
     QTRY_COMPARE(statusSpy.count(), 0);
 
     // Loading local file
     ctxt->setContextProperty("srcImage", testFileUrl("colors.gif"));
-    QTRY_VERIFY(obj->status() == QQuickImage::Ready);
-    QTRY_VERIFY(obj->progress() == 1.0);
+    QTRY_COMPARE(obj->status(), QQuickImage::Ready);
+    QTRY_COMPARE(obj->progress(), 1.0);
     QTRY_COMPARE(sourceSpy.count(), 1);
     QTRY_COMPARE(progressSpy.count(), 0);
     QTRY_COMPARE(statusSpy.count(), 1);
 
     // Loading remote file
     ctxt->setContextProperty("srcImage", server.url("/stickman.gif"));
-    QTRY_VERIFY(obj->status() == QQuickImage::Loading);
-    QTRY_VERIFY(obj->progress() == 0.0);
-    QTRY_VERIFY(obj->status() == QQuickImage::Ready);
-    QTRY_VERIFY(obj->progress() == 1.0);
+    QTRY_COMPARE(obj->status(), QQuickImage::Loading);
+    QTRY_COMPARE(obj->progress(), 0.0);
+    QTRY_COMPARE(obj->status(), QQuickImage::Ready);
+    QTRY_COMPARE(obj->progress(), 1.0);
     QTRY_COMPARE(sourceSpy.count(), 2);
     QTRY_VERIFY(progressSpy.count() > 1);
     QTRY_COMPARE(statusSpy.count(), 3);
 
     ctxt->setContextProperty("srcImage", "");
-    QTRY_VERIFY(obj->status() == QQuickImage::Null);
-    QTRY_VERIFY(obj->progress() == 0.0);
+    QTRY_COMPARE(obj->status(), QQuickImage::Null);
+    QTRY_COMPARE(obj->progress(), 0.0);
     QTRY_COMPARE(sourceSpy.count(), 3);
     QTRY_VERIFY(progressSpy.count() > 2);
     QTRY_COMPARE(statusSpy.count(), 4);
@@ -471,8 +519,8 @@ void tst_qquickanimatedimage::playingAndPausedChanges()
     QQmlComponent component(&engine);
     component.setData(componentStr.toLatin1(), QUrl::fromLocalFile(""));
     QQuickAnimatedImage *obj = qobject_cast<QQuickAnimatedImage*>(component.create());
-    QVERIFY(obj != 0);
-    QVERIFY(obj->status() == QQuickAnimatedImage::Null);
+    QVERIFY(obj != nullptr);
+    QCOMPARE(obj->status(), QQuickAnimatedImage::Null);
     QTRY_VERIFY(obj->isPlaying());
     QTRY_VERIFY(!obj->isPaused());
     QSignalSpy playingSpy(obj, SIGNAL(playingChanged()));
@@ -537,8 +585,8 @@ void tst_qquickanimatedimage::noCaching()
     window_nocache.setSource(testFileUrl("colors_nocache.qml"));
     window.show();
     window_nocache.show();
-    QTest::qWaitForWindowExposed(&window);
-    QTest::qWaitForWindowExposed(&window_nocache);
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QVERIFY(QTest::qWaitForWindowExposed(&window_nocache));
 
     QQuickAnimatedImage *anim = qobject_cast<QQuickAnimatedImage *>(window.rootObject());
     QVERIFY(anim);
@@ -562,6 +610,60 @@ void tst_qquickanimatedimage::noCaching()
     }
 }
 
+void tst_qquickanimatedimage::sourceChangesOnFrameChanged()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("colors.qml"));
+    QVector<QQuickAnimatedImage*> images;
+
+    // Run multiple animations in parallel, this should be fast
+    for (int loops = 0; loops < 25; ++loops) {
+        QQuickAnimatedImage *anim = qobject_cast<QQuickAnimatedImage *>(component.create());
+
+        // QTBUG-67427: this should not produce a segfault
+        QObject::connect(anim,
+                         &QQuickAnimatedImage::frameChanged,
+                         [this, anim]() { anim->setSource(testFileUrl("hearts.gif")); });
+
+        QVERIFY(anim);
+        QVERIFY(anim->isPlaying());
+
+        images.append(anim);
+    }
+
+    for (auto *anim : images)
+        QTRY_COMPARE(anim->source(), testFileUrl("hearts.gif"));
+
+    qDeleteAll(images);
+}
+
+void tst_qquickanimatedimage::currentFrame()
+{
+    QQuickView window;
+    window.setSource(testFileUrl("currentframe.qml"));
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+
+    QQuickAnimatedImage *anim = qobject_cast<QQuickAnimatedImage *>(window.rootObject());
+    QVERIFY(anim);
+    QSignalSpy frameChangedSpy(anim, SIGNAL(frameChanged()));
+    QSignalSpy currentFrameChangedSpy(anim, SIGNAL(currentFrameChanged()));
+
+    anim->setCurrentFrame(1);
+    QCOMPARE(anim->currentFrame(), 1);
+    QCOMPARE(frameChangedSpy.count(), 1);
+    QCOMPARE(currentFrameChangedSpy.count(), 1);
+    QCOMPARE(anim->property("currentFrameChangeCount"), 1);
+    QCOMPARE(anim->property("frameChangeCount"), 1);
+
+    evaluate<void>(anim, "scriptedSetCurrentFrame(2)");
+    QCOMPARE(anim->currentFrame(), 2);
+    QCOMPARE(frameChangedSpy.count(), 2);
+    QCOMPARE(currentFrameChangedSpy.count(), 2);
+    QCOMPARE(anim->property("currentFrameChangeCount"), 2);
+    QCOMPARE(anim->property("frameChangeCount"), 2);
+}
+
 QTEST_MAIN(tst_qquickanimatedimage)
 
 #include "tst_qquickanimatedimage.moc"
